#!/usr/bin/python

program_name = "pyrpn"
program_version = "0.3"

cfg = None

#
# PyRPN
#
# Copyright (C) 2002 Scott Scriven
#
# Released under the terms of the GNU Public License
#

# GTK stuffs
# Takes a while to import gtk on some systems
print "Loading GTK..."
from gtk import *
print "Loading GTK...  done."

def runqueue():
  while events_pending():
    mainiteration()


######################################################################
# Config class:
#
def is_config(obj):
  return hasattr(obj,'is_config') and obj.is_config

class config:
  is_config = 1

  def __init__(self):
    import os
    self.options = {}
    self.usrfilename = "%s/.%s.conf" % (os.environ["HOME"], program_name)
    self.globalfilename = "/etc/%s/%s.conf" % (program_name, program_name)
    self.namespace = {
      "options" : self.options,
      }

  ##################################################
  #
  def load(self):
    import os.path
    for name in [self.globalfilename, self.usrfilename]:
      if os.path.exists(name):
        self.parse(name)

  ##################################################
  #
  def parse(self, filename):
    #print "parsing %s..." % filename
    
    try:
      execfile(filename, self.namespace)
    except:
      print "Error in config file %s" % filename
      import traceback, sys
      foo = traceback.format_exception(sys.exc_type,
                                       sys.exc_value,
                                       sys.exc_traceback)
      for item in foo:
        item = item[:-1]
        print item
      print "Press Enter to continue..."
      raw_input()



######################################################################
# Simple convenience object for multiple-choice user queries
#
class ButtonChoiceWindow:

  def __init__(self):
    pass

  def __call__(self, title, text, buttons, default=0, long=0,
               transient=None):
    self._clicked = None

    win = GtkWindow()
    if transient:
      win.set_transient_for(transient)
    win.connect("destroy", mainquit)
    win.set_position(WIN_POS_CENTER)
    win.set_modal(1)
    win.set_title(title)
    win.set_border_width(5)

    v = GtkVBox()

    #if len(text) < 300:
    if not long:

      l = GtkLabel(text)
      l.set_line_wrap(1)
      l.show()
      v.add(l)

    else:  #if len(text) > 300:

      lbox = GtkHBox()
      l = GtkText(None, None)
      l.set_word_wrap(1)
      l.set_line_wrap(1)
      scroll = GtkVScrollbar(l.get_vadjustment())
      lbox.add(l, fill=1, expand=1)
      lbox.add(scroll, fill=0, expand=0)
      l.insert(None, None, None, text)
      #l.set_usize(400, 300)
      v.add(lbox, fill=1, expand=1)


    box = GtkHButtonBox()
    box.set_border_width(10)


    buttonlist = []
    i = 0
    for item in buttons:
      if type(item) == type((1,2)):
        label, value = item
      else:
        label = value = item


      b = GtkButton(label)
      buttonlist.append([label, value, b])
      b.connect("clicked", self._cb)

      if default >= 0  and  i == default:
        b.set_flags(CAN_DEFAULT)
        b.set_flags(HAS_DEFAULT)
        box.add(b)
      else:
        f = GtkFrame()
        f.set_border_width(2)
        f.set_shadow_type(SHADOW_NONE)
        f.add(b)
        box.add(f)

      i = i + 1

    box.show()
    v.add(box, fill=0, expand=0)

    v.show()
    win.add(v)

    win.show_all()
    mainloop()

    retval = None

    for item in buttonlist:
      label, value, b = item
      if b == self._clicked:
        retval = value

    win.hide()
    win.unrealize()
    del win

    return retval

  def _cb(self, button):
    self._clicked = button
    mainquit()

##################################################
#
def ltoa(num, base):
  symbols = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  symbols = symbols[:base]

  text = ""
  div = long(base)
  while num > 0:
    mod = num % div
    text = symbols[mod] + text
    num = (num - mod) / div
  #text = "b%s: %s" % (base, repr(num))
  #text = text[:-1]
  # TODO: print in requested base
  #text = repr(num)
  #text = "base%s: %s" % (base, text)
  return text



######################################################################
#
class value:
  def __init__(self, data):
    # Actual value held by this object
    self.data = data
    # Displayable representation of object
    self.cached = ""
    # Datatype
    self.type = "generic"

  def str(self):
    if not self.cached:
      self.cached = repr(self.data)
    return self.cached

  def set(self, data):
    self.data = data
    self.cached = ""

######################################################################
#
class Int(value):
  def __init__(self, data):
    value.__init__(self, long(data))
    self.type = "int"
    # Number base to display in
    self.base = 10
    self.base_cached = 0

  def str(self):
    if not self.cached  or  self.base != self.base_cached:
      global cfg
      
      if self.base == 10:
        text = repr(self.data)
        text = text[:-1]
      else:
        text = ltoa(self.data, self.base)
      # add commas
      commas = cfg.options["commas"]
      comma = cfg.options["comma type"]
      if commas:
        i = len(text) - commas
        while i > 0:
          text = text[:i] + comma + text[i:]
          i = i - commas
      if self.base != 10:
        text = "base" + repr(self.base) + ": " + text
      self.cached = text
      self.base_cached = self.base
    return self.cached

######################################################################
#
class Float(value):
  def __init__(self, data):
    value.__init__(self, float(data))
    self.type = "float"
    # TODO? set precision on a per-number basis?

  def str(self):
    if not self.cached:
      global cfg
      if cfg.options["precision"] > 0:
        format = "%%.%if" % (cfg.options["precision"])
      else:
        format = "%s"
      text = format % (self.data)

      # add commas
      commas = cfg.options["commas"]
      comma = cfg.options["comma type"]
      if commas:
        import string
        i = string.rfind(text, ".") - commas
        while i > 0:
          text = text[:i] + comma + text[i:]
          i = i - commas
      self.cached = text
    return self.cached

######################################################################
#
class Str(value):
  def __init__(self, data):
    value.__init__(self, data)
    self.type = "str"



######################################################################
#
# TODO: make sure all functions are "undo-safe"
#
# Key:
#  - Not done
#  + Half done
#  * Done
# 
# Basic:
#  * +, -, *, /
#  * modulus
#  - divmod
#  * sin, cos, tan
#  * asin, acos, atan
#  * sinh, cosh, tanh
#  + asinh, acosh, atanh
#  * sqrt, y^x, 1/x; x^2, xroot, 10^x, log, e^x, ln
#  * min / max
#  * abs, sign
#  - ipart, fpart
#  * floor, ceil
#  * rand, frand
#
# Integer / Logic:
#  * ->base (num, from, to) (up to base 36 (0-9,A-Z))
#  + lshift, rshift, rroll, lroll
#  - atonl, ntoal
#  + and, or, xor, not
#  + &, |, ^, bnot
#  + ->int, ->float
#
# Stack:
#  * last / undo
#  * drop
#  * clear
#  * dup
#  - dupn, dropn
#  * swap
#  * roll
#  + rolln (roll by x or -x items)
#  * pick
#  * edit
#
# Vector:
#  - arrays, lists, matrices
#
# Complex:
#  - complex numbers...
#  - i
#
# Symbolic:
#  - fractions
#  - equations
#  - value plugger
#
# Datatypes, units:
#  * explicit datatyping (type, value, units)
#  - datatype conversion (->int, etc)
#  * cached displayable values ?
#  - unit conversion
#    - deg, rad
#    - inch, foot, yard, mile, mm, cm, m, km
#    - g, kg, lbs, ...
#    - ...
#
# Algebra:
#  - eval expressions
#
# Statistics / probability:
#  + combination
#  + permutation
#  * !
#
# Financial
#
# Constants:
#  - e, i, pi, symbolic and numeric versions of each?
#
# Storage:
#  - directories, filesystem, etc...
#  - sto
#  - purge
#
# Programming:
#  - <<, >>
#  - if
#  - for
#  - while
#
# Options / modes:
#  * precision
#  + degrees, radians, grads?
#  + std (precision 0) / fixed (precision n) / sci floats
#
class stack:

  def __init__(self, cfg, parent=None):
    self.data = []
    self.parent = parent
    self.cfg = cfg

    self.functions = {
      "dump" : self.dump,
      
      "undo" : self.undo,
      "drop" : self.drop,
      "swap" : self.swap,
      "dup" : self.dup,
      "roll" : self.roll,
      "rolln" : self.rolln,
      "pick" : self.pick,
      "clear" : self.clear,
      "edit" : self.edit,
      
      "+" : self.add,
      "-" : self.subtract,
      "*" : self.multiply,
      "/" : self.divide,
      "neg" : self.negate,
      "%" : self.modulus,

      "min" : self.min,
      "max" : self.max,
      "abs" : (self.maths, ["abs"]),
      "sign" : self.sign,
      "floor" : (self.maths, ["floor"]),
      "ceil" : (self.maths, ["ceil"]),

      "rand" : self.rand,
      "frand" : self.frand,

      "perm" : self.perm,
      "comb" : self.comb,
      "!" : self.factorial,

      "->int" : self.to_int,
      "->float" : self.to_float,

      "shl" : self.shl,
      "shr" : self.shr,
      #"rol" : self.rol,
      #"ror" : self.ror,
      "&" : self.band,
      "|" : self.bor,
      "^" : self.bxor,
      "bnot" : self.bnot,
      "and" : self.iand,
      "or" : self.ior,
      "xor" : self.ixor,
      "not" : self.inot,
      
      "sqrt" : (self.maths, ["sqrt"]),
      "square" : self.square,
      "pow" : self.pow,
      "inv" : self.inv,
      "xroot" : self.xroot,
      "log" : (self.maths, ["log"]),
      "10pow" : self.pow10,
      "ln" : (self.maths, ["ln"]),
      "epow" : self.epow,
      
      "sin" : (self.maths, ["sin"]),
      "cos" : (self.maths, ["cos"]),
      "tan" : (self.maths, ["tan"]),
      "asin" : (self.maths, ["asin"]),
      "acos" : (self.maths, ["acos"]),
      "atan" : (self.maths, ["atan"]),
      "sinh" : (self.maths, ["sinh"]),
      "cosh" : (self.maths, ["cosh"]),
      "tanh" : (self.maths, ["tanh"]),
      "asinh" : (self.maths, ["asinh"]),
      "acosh" : (self.maths, ["acosh"]),
      "atanh" : (self.maths, ["atanh"]),
      
      "angle_unit" : self.angle_unit,
      "precision" : self.precision,
      "commas" : self.commas,
      "commatype" : self.comma_type,
      "base" : self.base,
      "->base" : self.to_base,
      }

  ##################################################
  #
  def execute(self, func):
    try:
      if func != self.undo:
        self.backup()
      if type(func) == type((1,2)):
        f = func[0]
        args = func[1]
        apply(f, args)
      else:
        func()
    except Exception:
      import sys
      ei = sys.exc_info()
      text = "%s: %s" % (ei[0], ei[1])
      text = text[11:]
      self.error("%s" % text)

  ##################################################
  #
  def operate(self, func, force=0):

    if self.functions.has_key(func)  or  force:
      self.execute(self.functions[func])

    else:
      self.push(func)

  ##################################################
  #
  def value(self, data):
    if type(data) in [type(1L), type(1)]:
      return Int(data)
    elif type(data) == type(1.0):
      return Float(data)
    elif type(data) == type("abc"):
      return Str(data)
    else:
      return value(data)

  ##################################################
  #
  def invalidate_cache(self):
    for d in self.data:
      d.cached = ""

  ##################################################
  #
  def dump(self):
    i = len(self.data) - 1
    while i >= 0:
      print "%s: %s (%s)" % (i+1, self.data[i].str(), self.data[i].data)
      i = i - 1

  ##################################################
  # Return stack to previous state
  # (only works once)
  def backup(self):
    self.old_data = self.data[:]

  ##################################################
  # Return stack to previous state
  # (only works once)
  def undo(self):
    foo = self.old_data
    self.old_data = self.data
    self.data = foo

  ##################################################
  #
  def push(self, val):
    self.backup()
    data = self.value(val)
    self.data.insert(0, data)

  ##################################################
  #
  def error(self, text):
    title = "Stack Error"
    buttons = [("D'oh!", "ok")]
    ButtonChoiceWindow()(title, text, buttons, transient=self.parent)

  ##################################################
  #
  def drop(self):
    del self.data[0]

  ##################################################
  #
  def swap(self):
    a = self.data[0]
    b = self.data[1]
    self.data[0] = b
    self.data[1] = a
      
  ##################################################
  #
  def dup(self):
    a = self.data[0]
    self.push(a.data)
      
  ##################################################
  #
  def roll(self):
    self.rolln(1)
      
  ##################################################
  #
  def rolln(self, a=0):
    if not a:
      a = self.data[0]
      del self.data[0]

    l = len(self.data)
    if a >= l  or  a <= -l:
      raise ValueError, "Requested roll too large"

    if a > 0:
      slice = self.data[:a]
      del self.data[:a]
      self.data.extend(slice)
    else:
      slice = self.data[a:]
      del self.data[a:]
      self.data[:0] = slice
      
  ##################################################
  #
  def clear(self):
    self.data = []

  ##################################################
  #
  def edit(self):
    a = self.data[0]

    if a.type == "int":
      text = repr(a.data)
      text = text[:-1]
    elif a.type == "float":
      text = repr(a.data)
    elif a.type == "str":
      text = a.data
    else:
      text = repr(a.data)

    title = "%s (edit)" % (program_name)
    import pyrpn_editor
    edit = pyrpn_editor.Editor(cfg, None, text, title, "small")
    edit.run()
    text = edit.get_text()

    if a.type == "int":
      v = Int(text)
    elif a.type == "float":
      v = Float(text)
    elif a.type == "str":
      v = Str(text)
    else:
      v = value(text)

    self.data[0] = v
      
  ##################################################
  #
  def pick(self):
    a = self.data[0].data
    b = self.data[a]
    del self.data[a]
    self.data[0] = b


  ##################################################
  #
  def add(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(b.data + a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def subtract(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(b.data - a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def negate(self):
    a = self.data[0]
    a.set(-a.data)
    self.data[0] = a

  ##################################################
  #
  def multiply(self):
    a = self.data[0]
    b = self.data[1]

    # coerce floats to ints for string multiplication
    if a.type == "str":
      if b.type == "float":
        b.data = int(b.data)
    elif b.type == "str":
      if a.type == "float":
        a.data = int(a.data)
        
    c = self.value(b.data * a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def divide(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(b.data / a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def modulus(self):
    a = self.data[0]
    b = self.data[1]
    if b.type == "float"  or  a.type == "float":
      import math
      c = Float(math.fmod(b.data, a.data))
      del self.data[1]
      self.data[0] = c
    elif b.type == "int"  and  a.type =="int":
      c = Int(b.data % a.data)
      del self.data[1]
      self.data[0] = c
    else:
      raise ValueError, "Invalid types for modulus"

  ##################################################
  #
  def min(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(min(a.data,b.data))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def max(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(max(a.data,b.data))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def sign(self):
    a = self.data[0]
    if a.data < 0:
      self.data[0] = Int(-1)
    else:
      self.data[0] = Int(1)

  ##################################################
  #
  def rand(self):
    import whrandom
    a = self.data[0]
    b = self.data[1]
    c = Int(whrandom.randint(b.data,a.data))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def frand(self):
    import whrandom
    self.push(whrandom.random())

  ##################################################
  #
  def perm(self):
    r = self.data[0]
    n = self.data[1]
    if n.data < r.data:
      raise ValueError, "Bad argument value"
    v = self._factorial(n.data) / self._factorial(n.data - r.data)
    del self.data[1]
    self.data[0] = Int(v)

  ##################################################
  #
  def comb(self):
    r = self.data[0]
    n = self.data[1]
    if n.data < r.data:
      raise ValueError, "Bad argument value"
    v = self._factorial(n.data) / \
        self._factorial(n.data - r.data) / \
        self._factorial(r.data)
    del self.data[1]
    self.data[0] = Int(v)

  ##################################################
  #
  def factorial(self):
    a = self.data[0]
    b = self._factorial(a.data)
    self.data[0] = Int(b)

  ##################################################
  #
  def _factorial(self, num):
    if not num > 0:
      raise ValueError, "Factorial takes a positive integer"
    b = 1L
    for c in xrange(num):
      b = b * (c + 1)
    return b

  ##################################################
  #
  def to_int(self):
    a = self.data[0]
    b = Int(a.data)
    self.data[0] = b

  ##################################################
  #
  def to_float(self):
    a = self.data[0]
    b = Float(a.data)
    self.data[0] = b

  ##################################################
  #
  def shl(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(b.data << a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def shr(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(b.data >> a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def band(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(a.data & b.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def bor(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(a.data | b.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def bxor(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(a.data ^ b.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def bnot(self):
    a = self.data[0]
    # FIXME: how can I do a binary not?
    b = -1L
    c = Int(a.data ^ b)
    self.data[0] = c

  ##################################################
  #
  def iand(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(not(not(a.data and b.data)))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def ior(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(not(not(a.data or b.data)))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def ixor(self):
    a = self.data[0]
    b = self.data[1]
    c = 0
    if (a.data and (not b.data)) or \
       (b.data and (not a.data)):
      c = 1
    c = Int(c)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def inot(self):
    a = self.data[0]
    c = Int(not a.data)
    self.data[0] = c

  ##################################################
  # Execute simple one-argument functions from the
  # math module...
  # a(sin|cos|tan)h seem to be missing
  #
  def maths(self, which):
    import math
    trig = ["sin", "cos", "tan",
            "sinh", "cosh", "tanh"]
    atrig = ["asin", "acos", "atan",
            "asinh", "acosh", "atanh"]
    funcs = {
      "sin" : math.sin,
      "cos" : math.cos,
      "tan" : math.tan,
      "asin" : math.asin,
      "acos" : math.acos,
      "atan" : math.atan,
      "sinh" : math.sinh,
      "cosh" : math.cosh,
      "tanh" : math.tanh,
      #"asinh" : math.asinh,
      #"acosh" : math.acosh,
      #"atanh" : math.atanh,
      "sqrt" : math.sqrt,
      "log" : math.log10,
      "ln" : math.log,
      "abs" : abs,
      "floor" : math.floor,
      "ceil" : math.ceil,
      }
    a = self.data[0]
    if (self.cfg.options["angle unit"] == "degrees")  and  (which in trig):
      a.set(a.data * math.pi / 180.0)
    func = funcs[which]
    a.set(func(a.data))
    if (self.cfg.options["angle unit"] == "degrees")  and  (which in atrig):
      a.set(a.data * 180.0 / math.pi)
    self.data[0] = self.value(a.data)
      

  ##################################################
  #
  def square(self):
    import math
    a = self.data[0]
    self.data[0] = self.value(a.data*a.data)

  ##################################################
  #
  def pow(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(b.data ** a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def inv(self):
    a = self.data[0]
    b = self.value(1 / a.data)
    self.data[0] = b

  ##################################################
  #
  def xroot(self):
    a = Float(self.data[0].data)
    b = self.data[1]
    a.set(b.data ** (1/a.data))
    del self.data[1]
    self.data[0] = a

  ##################################################
  #
  def epow(self):
    import math
    a = self.data[0]
    self.data[0] = Float(math.e**a.data)

  ##################################################
  #
  def pow10(self):
    a = self.data[0]
    self.data[0] = Float(10.0**a.data)

  ##################################################
  #
  def angle_unit(self):
    a = self.data[0]
    if a.data in ["degrees", "deg"]:
      self.cfg.options["angle unit"] = "degrees"
    elif a.data in ["radians", "rad"]:
      self.cfg.options["angle unit"] = "radians"
    else:
      raise ValueError, "Invalid angle unit"

    del self.data[0]

  ##################################################
  #
  def precision(self):
    a = self.data[0]
    a = int(a.data)
    if a < 0:
      raise ValueError, "Precision must be non-negative integer"
    self.cfg.options["precision"] = a

    del self.data[0]
    self.invalidate_cache()

  ##################################################
  #
  def commas(self):
    a = self.data[0]
    a = int(a.data)
    if a < 0:
      raise ValueError, "Commas must be non-negative integer"
    self.cfg.options["commas"] = a

    del self.data[0]
    self.invalidate_cache()

  ##################################################
  #
  def comma_type(self):
    a = self.data[0]
    a = a.data
    self.cfg.options["comma type"] = a

    del self.data[0]
    self.invalidate_cache()

  ##################################################
  #
  def base(self):
    a = self.data[0]
    a = int(a.data)
    if a < 2  or  a > 36:
      raise ValueError, "Base must be integer 2 ... 36"
    self.cfg.options["base"] = a

    del self.data[0]

  ##################################################
  #
  def to_base(self):
    a = self.data[0]
    b = self.data[1]
    a = int(a.data)
    if b.type != "int":
      raise ValueError, "Only integers can be converted to other bases"
    if a < 2  or  a > 36:
      raise ValueError, "Base must be integer 2 ... 36"
    b.base = a

    del self.data[0]

######################################################################
#
# misc todo:
# - insert commas in numbers > 3 digits
#
# Possible frame/tab configuration:
# - basic: numpad, main ops
# - stack
# - alpha (fitaly)
# - trig
# - programming
# - matrix
# - units
# - file manager (always visible?)
#
class calculator:

  def __init__(self, cfg):
    self.cfg = cfg

    self.win = GtkWindow()
    self.win.connect("destroy", self.destroy)
    self.win.set_title("%s %s" % (program_name, program_version))
    self.win.set_default_size(self.cfg.options["width"],
                              self.cfg.options["height"])

    self.stack = stack(self.cfg, self.win)

    self.vbox = GtkVBox()
    self.win.add(self.vbox)

    self.make_menus()
    self.make_stacklist()
    self.make_tophalf()
    self.make_bottomhalf()


  ##################################################
  #
  def make_menus(self):
    self.ag = GtkAccelGroup()
    self.win.add_accel_group(self.ag)
    self.itemf = GtkItemFactory(GtkMenuBar, "<main>", self.ag)
    foo = [("/_Kill", None, None, 0, "<Branch>")]
    self.itemf.create_items([
      ("/File", None,  None,  0,  "<Branch>"),
      ("/File/_Quit", "<Control>Q",  self.destroy,  0,  ""),
      ("/Help", None,  None,  0,  "<Branch>"),
      ("/Help/About", None,  self.about_screen,  0,  ""),
      ("/Help/Functions", None,  self.functions_screen,  0,  ""),
      ])
    self.menubar = self.itemf.get_widget("<main>")
    self.vbox.pack_start(self.menubar, fill=1, expand=0)

    foo = GtkItemFactory(GtkMenuBar, "<keydefs>", self.ag)
    foo.create_items([
      ("/foo",       None, None, 0, "<Branch>"),
      #("/foo/quit",  "<Control>Q", self.destroy, 0, ""),
      ])


  ##################################################
  #
  def make_stacklist(self):
    sw = GtkScrolledWindow()
    sw.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
    sw.set_placement(CORNER_BOTTOM_RIGHT)

    self.stack_lb = GtkCList(2)
    #self.stack_lb.connect("select-row", self._stacklist_select_cb)
    self.stack_lb.set_selection_mode(SELECTION_BROWSE)
    self.stack_lb.set_column_justification(0, JUSTIFY_RIGHT)
    self.stack_lb.set_column_justification(1, JUSTIFY_RIGHT)
    self.stack_lb.set_column_auto_resize(0, 1)
    self.stack_lb.set_column_auto_resize(1, 1)
    sw.add(self.stack_lb)
    
    self.vbox.pack_start(sw, fill=TRUE, expand=TRUE)

    self.entry = GtkEntry()
    self.entry.connect("activate", self._run_entry)
    self.vbox.add(self.entry, fill=0, expand=0)

  ##################################################
  #
  def update_stack(self):
    self.stack_lb.freeze()
    self.stack_lb.clear()
    for i in xrange(len(self.stack.data)):
      val = self.stack.data[i]
      text = val.str()
      self.stack_lb.insert(0, ["%s:" % (i), text])

    self.stack_lb.thaw()

    # scrollbar needs to automatically scroll to the bottom
    a = self.stack_lb.get_vadjustment()
    if a.upper > a.page_size:
      scrollto = a.upper - a.page_size
      a.set_value(scrollto)
    self.entry.grab_focus()

  ##################################################
  #
  def make_tophalf(self):

    nb = GtkNotebook()
    self.vbox.add(nb)

    t = self.make_exp()
    nb.append_page(t, GtkLabel("exp"))

    t = self.make_trig()
    nb.append_page(t, GtkLabel("trig"))

    #t = self.make_val()
    #nb.append_page(t, GtkLabel("val"))

    t = self.make_stat()
    nb.append_page(t, GtkLabel("stat"))

    t = self.make_int()
    nb.append_page(t, GtkLabel("int"))

    t = self.make_modes()
    nb.append_page(t, GtkLabel("modes"))


  ##################################################
  #
  def make_trig(self):
    t = GtkTable(2,6)
    x = 0
    
    b = GtkButton("sin")
    b.connect("clicked", self._operate, "sin")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("asin")
    b.connect("clicked", self._operate, "asin")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = GtkButton("cos")
    b.connect("clicked", self._operate, "cos")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("acos")
    b.connect("clicked", self._operate, "acos")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("tan")
    b.connect("clicked", self._operate, "tan")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("atan")
    b.connect("clicked", self._operate, "atan")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("sinh")
    b.connect("clicked", self._operate, "sinh")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("(asinh)")
    b.connect("clicked", self._operate, "asinh")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("cosh")
    b.connect("clicked", self._operate, "cosh")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("(acosh)")
    b.connect("clicked", self._operate, "acosh")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("tanh")
    b.connect("clicked", self._operate, "tanh")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("(atanh)")
    b.connect("clicked", self._operate, "atanh")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    return t
  
  ##################################################
  #
  def make_exp(self):
    t = GtkTable(2,5)
    x = 0
    
    b = GtkButton("x^2")
    b.connect("clicked", self._operate, "square")
    t.attach(b, 0, 1, 0, 1)
    b = GtkButton("sqrt")
    b.connect("clicked", self._operate, "sqrt")
    t.attach(b, 0, 1, 1, 2)
    x = x + 1
    
    b = GtkButton("y^x")
    b.connect("clicked", self._operate, "pow")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("xroot")
    b.connect("clicked", self._operate, "xroot")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("log")
    b.connect("clicked", self._operate, "log")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("10^x")
    b.connect("clicked", self._operate, "10pow")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("ln")
    b.connect("clicked", self._operate, "ln")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("e^x")
    b.connect("clicked", self._operate, "epow")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("1/x")
    b.connect("clicked", self._operate, "inv")
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    return t
  
  ##################################################
  #
  def make_val(self):
    t = GtkTable(2,4)
    
    return t
  
  ##################################################
  #
  def make_stat(self):
    t = GtkTable(2,7)
    x = 0
    
    b = GtkButton("!")
    b.connect("clicked", self._operate, "!")
    t.attach(b, x, x+1, 0, 1)
    #b = GtkButton("max")
    #b.connect("clicked", self._operate, "max")
    #t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("rand")
    b.connect("clicked", self._operate, "rand")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("frand")
    b.connect("clicked", self._operate, "frand")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = GtkButton("perm")
    b.connect("clicked", self._operate, "perm")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("comb")
    b.connect("clicked", self._operate, "comb")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("mod")
    b.connect("clicked", self._operate, "%")
    t.attach(b, x, x+1, 0, 1)
    # divmod?
    #b = GtkButton("sqrt")
    #b.connect("clicked", self._operate, "sqrt")
    #t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = GtkButton("max")
    b.connect("clicked", self._operate, "max")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("min")
    b.connect("clicked", self._operate, "min")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("abs")
    b.connect("clicked", self._operate, "abs")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("sign")
    b.connect("clicked", self._operate, "sign")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("ceil")
    b.connect("clicked", self._operate, "ceil")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("floor")
    b.connect("clicked", self._operate, "floor")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    return t
  
  ##################################################
  #
  def make_int(self):
    t = GtkTable(2,8)
    x = 0
    
    b = GtkButton("->base")
    b.connect("clicked", self._operate, "->base")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("base")
    b.connect("clicked", self._operate, "base")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("->int")
    b.connect("clicked", self._operate, "->int")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("->float")
    b.connect("clicked", self._operate, "->float")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("shl")
    b.connect("clicked", self._operate, "shl")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("shr")
    b.connect("clicked", self._operate, "shr")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = GtkButton("(rol)")
    b.connect("clicked", self._operate, "rol")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("(ror)")
    b.connect("clicked", self._operate, "ror")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("&")
    b.connect("clicked", self._operate, "&")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("|")
    b.connect("clicked", self._operate, "|")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = GtkButton("^")
    b.connect("clicked", self._operate, "^")
    t.attach(b, x, x+1, 0, 1)
    #b = GtkButton("bnot")
    #b.connect("clicked", self._operate, "bnot")
    #t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = GtkButton("and")
    b.connect("clicked", self._operate, "and")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("or")
    b.connect("clicked", self._operate, "or")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = GtkButton("xor")
    b.connect("clicked", self._operate, "xor")
    t.attach(b, x, x+1, 0, 1)
    b = GtkButton("not")
    b.connect("clicked", self._operate, "not")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    #b = GtkButton("aton")
    #b.connect("clicked", self._operate, "aton")
    #t.attach(b, x, x+1, 0, 1)
    #b = GtkButton("ntoa")
    #b.connect("clicked", self._operate, "ntoa")
    #t.attach(b, x, x+1, 1, 2)
    #x = x + 1

    return t
  
  ##################################################
  #
  def make_modes(self):
    t = GtkTable(2,4)
    x = 0

    b = GtkButton("precision")
    b.connect("clicked", self._operate, "precision")
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    b = GtkButton()
    text = "deg"
    if self.cfg.options["angle unit"] in ["rad", "radians"]:
      text = "rad"
    self._degrad_label = GtkLabel(text)
    b.add(self._degrad_label)
    b.connect("clicked", self._degrad)
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    b = GtkButton("base")
    b.connect("clicked", self._operate, "base")
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    b = GtkButton("commas")
    b.connect("clicked", self._operate, "commas")
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    return t



  ##################################################
  #
  def make_bottomhalf(self):
    nb = GtkNotebook()
    self.vbox.add(nb)

    numpad = self._make_numpad()
    nb.append_page(numpad, GtkLabel("num"))

    alphapad = self._make_alphapad()
    nb.append_page(alphapad, GtkLabel("alpha"))
    
    #grapher = GtkFrame()
    #nb.append_page(grapher, GtkLabel("graph"))
    
  ##################################################
  #
  def _make_numpad(self):
    hb = GtkHBox()
    hb.set_homogeneous(0)
    hb.set_spacing(10)

    t0 = GtkTable(3, 2)
    
    b = GtkButton("undo")
    b.connect("clicked", self._operate, "undo")
    t0.attach(b, 1, 2, 0, 1)

    b = GtkButton("swap")
    b.connect("clicked", self._operate, "swap")
    t0.attach(b, 0, 1, 0, 1)

    b = GtkButton("pick")
    b.connect("clicked", self._operate, "pick")
    t0.attach(b, 1, 2, 1, 2)

    b = GtkButton("roll")
    b.connect("clicked", self._operate, "roll")
    t0.attach(b, 0, 1, 1, 2)

    #b = GtkButton("rolln")
    #b.connect("clicked", self._operate, "roll")
    #t0.attach(b, 1, 2, 2, 3)

    b = GtkButton("edit")
    b.connect("clicked", self._operate, "edit")
    t0.attach(b, 1, 2, 2, 3)

    b = GtkButton("clear")
    b.connect("clicked", self._operate, "clear")
    t0.attach(b, 0, 1, 2, 3)

    hb.add(t0)
    
    
    t = GtkTable(4,3)
    t.set_homogeneous(1)
    
    b0 = GtkButton("0")
    bdot = GtkButton(".")
    b1 = GtkButton("1")
    b2 = GtkButton("2")
    b3 = GtkButton("3")
    b4 = GtkButton("4")
    b5 = GtkButton("5")
    b6 = GtkButton("6")
    b7 = GtkButton("7")
    b8 = GtkButton("8")
    b9 = GtkButton("9")
    
    t.attach(bdot, 1, 2, 3, 4)
    t.attach(b0, 0, 1, 3, 4)
    t.attach(b1, 0, 1, 2, 3)
    t.attach(b2, 1, 2, 2, 3)
    t.attach(b3, 2, 3, 2, 3)
    t.attach(b4, 0, 1, 1, 2)
    t.attach(b5, 1, 2, 1, 2)
    t.attach(b6, 2, 3, 1, 2)
    t.attach(b7, 0, 1, 0, 1)
    t.attach(b8, 1, 2, 0, 1)
    t.attach(b9, 2, 3, 0, 1)

    bdot.connect("clicked", self._type_into_entry, ".")
    b0.connect("clicked", self._type_into_entry, "0")
    b1.connect("clicked", self._type_into_entry, "1")
    b2.connect("clicked", self._type_into_entry, "2")
    b3.connect("clicked", self._type_into_entry, "3")
    b4.connect("clicked", self._type_into_entry, "4")
    b5.connect("clicked", self._type_into_entry, "5")
    b6.connect("clicked", self._type_into_entry, "6")
    b7.connect("clicked", self._type_into_entry, "7")
    b8.connect("clicked", self._type_into_entry, "8")
    b9.connect("clicked", self._type_into_entry, "9")

    bneg = GtkButton("+/-")
    bneg.connect("clicked", self._operate, "neg")
    t.attach(bneg, 2, 3, 3, 4)

    hb.add(t, fill=1, expand=1)

    t2 = GtkTable(4,2)
    t2.set_homogeneous(1)

    #t2.attach(GtkButton(), 0, 1, 0, 1)
    #t2.attach(GtkButton(), 1, 2, 0, 1)
    bdrop = GtkButton("<--")
    bdrop.connect("clicked", self._drop)
    t2.attach(bdrop, 0, 2, 0, 1)
    bplus = GtkButton("+")
    bplus.connect("clicked", self._operate, "+")
    t2.attach(bplus, 0, 1, 1, 2)
    bminus = GtkButton("-")
    bminus.connect("clicked", self._operate, "-")
    t2.attach(bminus, 0, 1, 2, 3)
    bmult = GtkButton("*")
    bmult.connect("clicked", self._operate, "*")
    t2.attach(bmult, 1, 2, 1, 2)
    bdiv = GtkButton("/")
    bdiv.connect("clicked", self._operate, "/")
    t2.attach(bdiv, 1, 2, 2, 3)
    bret = GtkButton("Enter")
    bret.connect("clicked", self._run_entry)
    t2.attach(bret, 0, 2, 3, 4)


    hb.add(t2)

    return hb


  ##################################################
  #
  def _make_alphapad(self):
    self._shifted = 0

    rows = self.cfg.options["keymap"]
    wid = len(rows[0])
    hgt = len(rows)
    
    t = GtkTable(hgt,wid)
    t.set_homogeneous(1)

    y = 0
    for row in rows:
      x = 0
      for l in row:
        if l != " ":
          b = GtkButton(l)
          b.connect("clicked", self._type, l)
          t.attach(b, x, x+1, y, y+1)
        x = x + 1
      y = y + 1

    b = GtkButton(" ")
    b.connect("clicked", self._type, " ")
    bpos = [b] ; bpos.extend(self.cfg.options["space 1 pos"])
    apply(t.attach, bpos)
    b = GtkButton(" ")
    b.connect("clicked", self._type, " ")
    bpos = [b] ; bpos.extend(self.cfg.options["space 2 pos"])
    apply(t.attach, bpos)

    self.shift_label1 = GtkLabel("shf")
    self.shift_label2 = GtkLabel("shf")
    b = GtkButton()
    b.add(self.shift_label1)
    b.connect("clicked", self._type, "shift")
    bpos = [b] ; bpos.extend(self.cfg.options["shift 1 pos"])
    apply(t.attach, bpos)
    b = GtkButton()
    b.add(self.shift_label2)
    b.connect("clicked", self._type, "shift")
    bpos = [b] ; bpos.extend(self.cfg.options["shift 2 pos"])
    apply(t.attach, bpos)
    
    b = GtkButton("<--")
    b.connect("clicked", self._drop)
    bpos = [b] ; bpos.extend(self.cfg.options["<-- pos"])
    apply(t.attach, bpos)
    
    b = GtkButton("TAB")
    b.connect("clicked", self._type, "\t")
    bpos = [b] ; bpos.extend(self.cfg.options["tab pos"])
    apply(t.attach, bpos)

    b = GtkButton("RET")
    b.connect("clicked", self._run_entry)
    bpos = [b] ; bpos.extend(self.cfg.options["ret pos"])
    apply(t.attach, bpos)

    return t


  ##################################################
  #
  def _operate(self, b, func):
    # Automatically "press enter" if entry is not empty
    t = self.entry.get_text()
    if t != "":
      self.entry.activate()

    self.stack.operate(func)
    self.update_stack()

  ##################################################
  #
  def _type_into_entry(self, b, text):
    t = self.entry.get_text()
    p = self.entry.get_position()
    t = "%s%s%s" % (t[:p], text, t[p:])
    self.entry.set_text(t)
    self.entry.set_position(p+len(text))
    self.entry.grab_focus()

  ##################################################
  #
  def _type(self, b, text):
    if self._shifted:
      caps = self.cfg.options["caps"]
      if caps.has_key(text):
        text = caps[text]

    if len(text) > 1:
      if text == "shift":
        if self._shifted == 0:
          self._shift(1)
        elif self._shifted == 1:
          self._shift(2)
        else:
          self._shift(0)
    else:
      self._type_into_entry(b, text)
      if self._shifted == 1:
        self._shift(0)

  ##################################################
  #
  def _shift(self, val):
    self._shifted = val
    if val == 0:
      self.shift_label1.set_text("shf")
      self.shift_label2.set_text("shf")
    elif val == 1:
      self.shift_label1.set_text("Shf")
      self.shift_label2.set_text("Shf")
    elif val > 1:
      self.shift_label1.set_text("SHF")
      self.shift_label2.set_text("SHF")

  ##################################################
  #
  def _run_entry(self, *args):
    text = self.entry.get_text()
    if not text:  text = "dup"

    # Attempt to convert into a non-text datatype
    try:
      if "." in text:
        text = float(text)
      else:
        text = long(text, self.cfg.options["base"])
    except:
      pass

    # Execute command
    self.stack.operate(text)

    self.entry.set_text("")
    self.update_stack()

  ##################################################
  #
  # TODO: make this behave better at beginning/end of line
  def _entry_backspace(self, *args):
    t = self.entry.get_text()
    p = self.entry.get_position()
    if t != ""  and  p > 0:
      t = t[:p-1] + t[p:]
      self.entry.set_text(t)
      self.entry.set_position(p-1)
      self.entry.grab_focus()

  ##################################################
  #
  def _degrad(self, b, *args):
    unit = "degrees"
    l = self._degrad_label
    l.set_text("deg")
    if self.cfg.options["angle unit"] == "degrees":
      unit = "radians"
      l.set_text("rad")
    self._operate(None, unit)
    self._operate(None, "angle_unit")

  ##################################################
  #
  def _drop(self, *args):
    t = self.entry.get_text()
    if t != "":
      self._entry_backspace()
    else:
      self._operate(None, "drop")

  ##################################################
  #
  def about_screen(self, *args):
    text = \
"""%s V%s

Copyright (C) 2002 Scott Scriven <%s@xyzz.org>

This program is licened under the terms of the GNU General Public License (GPL), version 2.0 or later, as published by the Free Software Foundation.  See <http://www.gnu.org/copyleft/gpl.txt> for the terms of the latest version of the GNU General Public License.
""" % (program_name, program_version, program_name)

    ButtonChoiceWindow()("About", text, ["Ok"], long=1, transient=self.win)
    
  ##################################################
  #
  def functions_screen(self, *args):
    funcs = self.stack.functions.keys()
    funcs.sort()
    text = \
"""Function List
--------

You may execute commands either by clicking on buttons or by typing the name of the command into the stack.  Anything not recognized as a function will be pushed onto the stack.

The following %s functions are available:
""" % (len(funcs))
    for f in funcs:
      text = text + "\t%s\n" % (f)

    ButtonChoiceWindow()("Functions", text, ["Ok"], long=1, transient=self.win)
    
  ##################################################
  #
  def run(self):

    self.win.show_all()
    self.update_stack()
    mainloop()

  ##################################################
  #
  def destroy(self, *args):
    self.win.hide()
    self.win.unrealize()
    mainquit()


######################################################################
#
def run(args):
  global cfg
  cfg = config()
  cfg.load()

  c = calculator(cfg)
  c.run()


######################################################################
#
if __name__ == "__main__":
  import sys
  run(sys.argv[1:])
