from gtk import *

######################################################################
#
class EditWidget(GtkText):
  def __init__(self):
    GtkText.__init__(self)

    self.linewrap = 1
    self.set_editable(1)
    self.set_word_wrap(1)
    self.set_line_wrap(self.linewrap)
    #self.text = ""
    self.font = None

  ##################################################
  #
  def set_text(self, text):
    #print "set_text: %s" % (text)
    #self.text = text
    self.freeze()
    self.clear()
    self.insert(self.font, None, None, text)
    #self.insert_defaults(text)
    self.thaw()
    self.show()

  ##################################################
  #
  def get_text(self):
    #self.text = self.get_chars(0, self.get_length())
    return self.get_chars(0, self.get_length())
 

  ##################################################
  #
  def clear(self):
    self.set_point(0)
    self.forward_delete(self.get_length())

  ##################################################
  #
  def set_font(self, fontname):
    point = self.get_position()
    self.freeze()
    try:
	self.font = load_font(fontname)
    except:
	self.thaw()
	return
    style = self.get_style()
    style.font = self.font
    self.set_style(style)
    #self.thaw()
    #self.freeze()
    self.set_text(self.get_text())
    self.set_position(point)
    self.thaw()

  ##################################################
  #
  def toggle_linewrap(self):
    self.linewrap = self.linewrap ^ 1
    self.set_line_wrap(self.linewrap)
    

######################################################################
#
class Editor:
  def __init__(self, cfg, parent=None, text=None, title=None, style="full"):
    # styles: small, medium, full
    self.cfg = cfg
    self.title = title
    self.style = style
    self.filename = ""
    self._cwd = "."

    self.win = GtkWindow()
    win = self.win
    win.connect("delete_event", self._quit)
    win.set_modal(1)
    win.set_title("%s" % (self.title))
    
    self.vbox = GtkVBox()
    win.add(self.vbox)
    
    self._make_menus()
    
    t = GtkTable(2,2)
    t.set_row_spacing(0,2)
    t.set_col_spacing(0,2)
    self.vbox.pack_start(t, fill=TRUE, expand=TRUE)
    
    self.editor_widget = EditWidget()
    editor = self.editor_widget
    self.fontnum = 0
    # Set font first to "fixed", to work around a GTK bug.
    # If the requested font fails, settng font later alters *everything*
    # So, make sure the first font request doesn't fail.
    editor.set_font("fixed")
    editor.set_font(self.cfg.options["editor fonts"][0])

    if text: editor.set_text(text)


    hscroll = GtkHScrollbar(editor.get_hadjustment())
    vscroll = GtkVScrollbar(editor.get_vadjustment())

    t.attach(editor, 0,1, 0,1)
    t.attach(hscroll, 0,1, 1,2, yoptions=FILL)
    t.attach(vscroll, 1,2, 0,1, xoptions=FILL)
    
    if parent:
      win.set_transient_for(parent)
      print "Parent: %s" % parent
    # FIXME: size doesn't work!
    print "size: %sx%s" % (self.cfg.options["editor %s width" % (self.style)],
                         self.cfg.options["editor %s height" % (self.style)])
    #self.win.set_usize(self.cfg.options["editor %s width" % (self.style)],
    #                     self.cfg.options["editor %s height" % (self.style)])
    editor.set_usize(self.cfg.options["editor %s width" % (self.style)],
                         self.cfg.options["editor %s height" % (self.style)])
    win.show_all()
    
  ##################################################
  #
  def _make_menus(self):
    self.ag = GtkAccelGroup()
    self.win.add_accel_group(self.ag)
    self.itemf = GtkItemFactory(GtkMenuBar, "<main>", self.ag)
    self.itemf.create_items([
      #("/_File",       None, None, 0, "<Branch>"),
      #("/File/_Open",  "<Control>O", self._file_open, 0, ""),
      #("/File/_Reload",  "", self._reload, 0, ""),
      #("/File/_Save",  "<Control>S", self._save, 0, ""),
      #("/File/Save _As...",  None, self._save_as, 0, ""),
      #("/File/_Done",  "<Control>Q", self._quit, 0, ""),
      ("/_Edit",       None, None, 0, "<Branch>"),
      ("/Edit/_Cut",   None, self._cut, 0, ""),
      ("/Edit/_Copy",  None, self._copy, 0, ""),
      ("/Edit/_Paste", None, self._paste, 0, ""),
      ("/_Misc",       None, None, 0, "<Branch>"),
      ("/Misc/Change _Font",   None, self._change_font, 0, ""),
      ("/Misc/Toggle _Linewrap",   None, self._toggle_linewrap, 0, ""),
      ("/Done",        None, self._quit, 0, ""),
      #("/_About",       None, None, 0, "<Branch>"),
      #("/About/About %s" % (program_name), None, self.about_screen, 0, ""),
      ])
    self.menubar = self.itemf.get_widget("<main>")
    self.vbox.pack_start(self.menubar, fill=FALSE, expand=FALSE)
        
    # Extra keypresses to handle
    #foo = GtkItemFactory(GtkMenuBar, "<keydefs>", self.ag)
    #foo.create_items([
    #    ("/foo",       None, None, 0, "<Branch>"),
    #    ("/foo/play",  "P", self._play_button, 0, ""),
    #    ])


  ##################################################
  #
  def _cut(self, *args):
    self.editor_widget.cut_clipboard()

  ##################################################
  #
  def _copy(self, *args):
    self.editor_widget.copy_clipboard()

  ##################################################
  #
  def _paste(self, *args):
    self.editor_widget.paste_clipboard()

  ##################################################
  #
  def _toggle_linewrap(self, *args):
    self.editor_widget.toggle_linewrap()

  ##################################################
  #
  def _change_font(self, *args):
    self.fontnum = self.fontnum + 1
    if self.fontnum >= len(self.cfg.options["editor fonts"]):
      self.fontnum = 0
    self.editor_widget.set_font(self.cfg.options["editor fonts"][self.fontnum])

  ##################################################
  #
  def _file_open(self, *args):
    filename = self.file_selector("Open...", "")
    if filename:
      self.load(filename)
    
  ##################################################
  #
  def _reload(self, *args):
    if self.filename:
      self.load(self.filename)
    
  ##################################################
  #
  def _save(self, *args):
    if self.filename:
      self.save()
    else:
      self._save_as()

  ##################################################
  #
  def _save_as(self, *args):
    filename = self.file_selector("Save As...", "")
    if filename:
      self.save(filename)
    

  ##################################################
  #
  def file_selector(self, title, filename):
    import os.path
    self.file_selector_cb_button = None
    
    fs = GtkFileSelection(title)
    if filename:
      fs.set_filename(filename)
    else:
      fs.set_filename(self._cwd)
    fs.ok_button.connect("clicked", self.file_selector_ok_cb)
    fs.cancel_button.connect("clicked", self.file_selector_cancel_cb)
    fs.show_all()
    mainloop()

    if self.file_selector_cb_button == "ok":
      filename = fs.get_filename()
      self._cwd = "%s/" % (os.path.dirname(filename))
      fs.hide()
      return filename
    else:
      fs.hide()
      return None

  def file_selector_ok_cb(self, *args):
    #print "ok"
    self.file_selector_cb_button = "ok"
    mainquit()

  def file_selector_cancel_cb(self, *args):
    #print "cancel"
    self.file_selector_cb_button = "cancel"
    mainquit()


  ##################################################
  #
  def load(self, filename):
    print "Loading... %s" % (filename)
    import string
    self.filename = filename
    self.update_title()
    try:
      file = open(self.filename, "r")
      lines = file.readlines()
      text = string.join(lines, "")
      self.editor_widget.set_text(text)
      file.close()
      self.update_title()
    except IOError, msg:
      print "Cannot open file %s: %s" % (filename, msg)

  ##################################################
  #
  def save(self, filename=None):
    if filename:
      self.filename = filename
      self.update_title()
    if self.cfg.options["make backups"]:
      import shutil #os

      try:
        #os.rename(self.filename, "%s~" % (self.filename))
        shutil.copy(self.filename, "%s~" % (self.filename))
      except:
        pass

    try:
      file = open(self.filename, "w+")
      file.write(self.editor_widget.get_text())
      file.close()
    except Exception, msg:
      ButtonChoiceWindow("Error Saving",
                         "Could not save %s: %s" % (self.filename, msg),
                         ["Ok"])

  ##################################################
  #
  def update_title(self, *args):
    import string
    
    i = string.rfind(self.filename, "/")
    name = self.filename[i+1:]
    self.win.set_title("%s" % (name))


  ##################################################
  #
  def get_text(self):
    return self.editor_widget.get_text()

  ##################################################
  #
  def run(self):
    self.win.grab_focus()
    mainloop()

  ##################################################
  #
  def _quit(self, *args):
    self.quit()

  ##################################################
  #
  def quit(self):
    self.win.hide()
    self.win.unrealize()
    mainquit()


  ##################################################
  #
  def __del__(self):
    self.win.unrealize()
    del self.win


